import os
import pdb
import shutil

from .app import App
from .constants import COMPOSE_TEMPLATE, WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE
from .service_workflow import ServiceWorkflow
from .workflow_command import WorkflowCommand

CORE_WORKFLOWS = [WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE]

class WorkflowCopyCommand(WorkflowCommand):

  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

    self.__force = not not kwargs.get('force')

  @property
  def force(self):
    return self.__force

  def copy(self, destination_workflow_name: str):
    destination_workflow = ServiceWorkflow(self.service_name, destination_workflow_name, self.app)

    # Create workflow folder
    dest = os.path.join(destination_workflow.path)
    if os.path.exists(dest) and self.force:
        shutil.rmtree(dest)

    if not os.path.exists(dest):
      os.makedirs(dest) 

    self.app.copy_folders_and_hidden_files(self.workflow_path, dest)

    compose_file_src = os.path.join(dest, COMPOSE_TEMPLATE.format(workflow=self.workflow_name))
    compose_file_dest = os.path.join(dest, COMPOSE_TEMPLATE.format(workflow=destination_workflow_name))
    os.rename(compose_file_src, compose_file_dest)

    # Replace workflow name
    with open(compose_file_dest, 'r+') as fp:
      contents = fp.read()
      fp.seek(0)
      fp.write(contents.replace(self.workflow_name, destination_workflow_name))
      fp.truncate()


from typing import List, TypedDict
from urllib import request

from stoobly_agent.app.settings.parameter_rule import ParameterRule
from stoobly_agent.app.settings.constants import request_component

from ..mock.hashed_request_decorator import COMPONENT_TYPES

class IgnoredComponent(TypedDict):
  name: str
  type: str

def rewrite_rules_to_ignored_components(rules: List[ParameterRule]) -> List[IgnoredComponent]:
  ignored_components = []

  for rule in rules:
    rule_name = rule.name
    rule_type = rule.type

    if rule_type == request_component.HEADER:
      ignored_components.append({
        'name': rule_name,
        'type': COMPONENT_TYPES['HEADER']
      })
    elif rule_type == request_component.QUERY_PARAM:
      ignored_components.append({
        'name': rule_name,
        'type': COMPONENT_TYPES['QUERY_PARAM']
      })
    elif rule_type == request_component.BODY_PARAM:
      ignored_components.append({
        'name': rule_name,
        'type': COMPONENT_TYPES['BODY_PARAM']
      })
  
  return ignored_components



import click
import pdb
import sys

from stoobly_agent.app.settings import Settings

from .helpers import ProjectFacade
from .helpers.print_service import FORMATS, print_projects, select_print_options
from .helpers.validations import *

@click.group(
    epilog="Run 'stoobly-agent project COMMAND --help' for more information on a command.",
    help="Manage projects"
)
@click.pass_context
def project(ctx):
    pass

@project.command(
    help="Create a project"
)
@click.option('--description', help='Project description.')
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--organization-key', required=True, help='Project to create project in.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('name')
def create(**kwargs):
    print_options = select_print_options(kwargs)

    project = ProjectFacade(Settings.instance())

    try:
        project_response = project.create(**kwargs)
    except AssertionError as e:
        return print(e, file=sys.stderr)

    print_projects([project_response], **print_options)

@project.command(
    help="Show all projects"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--page', default=0)
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--size', default=10)
@click.option('--sort-by', default='created_at', help='created_at|name')
@click.option('--sort-order', default='desc', help='asc | desc')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('organization_key', required=False)
def list(**kwargs):
    print_options = select_print_options(kwargs)
    settings = Settings.instance()
    project = ProjectFacade(settings)

    organization_key = kwargs['organization_key']

    # If organization_key is not set, try to use current project's organization
    if not organization_key:
        try:
            project_key = ProjectKey(settings.proxy.intercept.project_key)

            if not project_key.is_local:
                organization_key = OrganizationKey.encode(project_key.organization_id)
        except InvalidProjectKey as e:
            pass

    # If organization_key is not set, try to use current user's private organization
    if not organization_key:
        try:
            user_profile = project.user_profile()
        except AssertionError as e:
            return print(e, file=sys.stderr)

        organization_key = user_profile['organization_key']

    validate_organization_key(organization_key)

    try:
        kwargs['organization_key'] = organization_key
        projects_response = project.index(kwargs)
    except AssertionError as e:
        return print(e, file=sys.stderr)

    if len(projects_response['list']) == 0:
        print('No projects found.', file=sys.stderr)
    else:
        print_projects(projects_response['list'], **print_options)

@project.command(
    help="Describe project"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.argument('project_key', required=False)
def show(**kwargs):
    print_options = select_print_options(kwargs)

    settings = Settings.instance()
    project_key = resolve_project_key_and_validate(kwargs, settings)
    project = ProjectFacade(settings)

    try:
        project_response = project.show(project_key)
    except AssertionError as e:
        return print(e, file=sys.stderr)

    print_projects([project_response], **print_options)


import pdb
import re
from typing import List

from mitmproxy.coretypes.multidict import MultiDict
from requests import request

from stoobly_agent.app.proxy.replay.rewrite_params_service import build_id_to_alias_map
from stoobly_agent.config.constants import test_filter
from stoobly_agent.lib.api.interfaces.endpoints import Alias, RequestComponentName
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias

class RequestComponentNamesFacade():

  def __init__(self, request_component_names_response):
    self.__request_component_names: List[RequestComponentName] = request_component_names_response

    self.__aliased_request_component_names = None
    self.__deterministic_request_component_names = None
    self.__required_request_component_names = None

    self.__filter: test_filter.TestFilter = test_filter.ALL
    self.__trace: Trace = None

    self.__request_component_name_map = {}
    for request_component_name in self.__request_component_names:
      if not request_component_name.get('query'):
        continue

      self.__request_component_name_map[request_component_name['query']] = request_component_name

    self.__id_to_alias_map = {}

    # Cache
    self.__edges_index_cache = None
    self.__query_index_cache = None

  @property
  def query_index(self):
    if self.__query_index_cache:
      return self.__query_index_cache

    _index = MultiDict()

    for request_component_name in self.__request_component_names:
      query = request_component_name.get('query')
      if not query:
        query = request_component_name.get('name')

      _index.add(query, request_component_name)

    self.__query_index_cache = _index

    return _index

  @property
  def edges_index(self):
    if self.__edges_index_cache:
      return self.__edges_index_cache

    _index = {}

    for request_component_name in self.__request_component_names:
      # TODO: make this less hacky, right now only response_param_names and body_param_names have children though
      parent_id = request_component_name.get('response_param_name_id')
      if not parent_id:
        parent_id = request_component_name.get('body_param_name_id')

      if parent_id not in _index:
        _index[parent_id] = []

      _index[parent_id].append(request_component_name)

    self.__edges_index_cache = _index 

    return _index

  @property
  def aliased(self) -> List[RequestComponentName]: 
    if not self.__aliased_request_component_names:
      filter_handler = lambda param_name: param_name.get('alias_id') != None
      self.__aliased_request_component_names = list(filter(filter_handler, self.__request_component_names))

    return self.__aliased_request_component_names

  @property
  def all(self) -> List[RequestComponentName]:
    return self.__request_component_names

  @property
  def deterministic(self) -> List[RequestComponentName]:
    if not self.__deterministic_request_component_names:
      filter_handler = lambda param_name: (not not param_name.get('is_deterministic') )
      self.__deterministic_request_component_names = list(filter(filter_handler, self.__request_component_names))

    return self.__deterministic_request_component_names

  @property
  def required(self) -> List[RequestComponentName]: 
    if not self.__required_request_component_names:
      filter_handler = lambda param_name: (not not param_name.get('is_required'))
      self.__required_request_component_names = list(filter(filter_handler, self.__request_component_names))

    return self.__required_request_component_names

  def is_selected(self, query: str) -> bool:
    if self.__filter == test_filter.ALL:
      return True
    elif self.__filter == test_filter.ALIAS:
      return self.__select_alias(query)
    elif self.__filter == test_filter.LINK:
      return self.__select_link(query)
    else:
      return False

  def with_aliases(self, aliases: List[Alias]):
    self.__id_to_alias_map = build_id_to_alias_map(aliases)
    return self

  def with_filter(self, filter: test_filter.TestFilter):
    self.__filter = filter
    return self

  def with_trace(self, trace: Trace):
    self.__trace = trace
    return self
    
  def __select_alias(self, query: str):
    _request_component_name = self.__request_component_name_map.get(query)
    if not _request_component_name:
      return False

    # If aliased, then filter
    return not not _request_component_name.get('alias_id')

  def __select_link(self, query: str):
    if not self.__trace:
      return True

    _request_component_name: RequestComponentName = self.__request_component_name_map.get(query)
    if not _request_component_name:
      return False

    alias_id = _request_component_name.get('alias_id')
    if not alias_id:
      return False

    _alias: Alias = self.__id_to_alias_map.get(alias_id)
    
    if not _alias:
      return False

    alias_name = _alias['name']
    trace_aliases = TraceAlias.where({
      'name': alias_name,
      'trace_id': self.__trace.id,
    })

    return trace_aliases.count() > 0


import click, sys

from stoobly_agent.app.models.types import OPENAPI_FORMAT
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.logger import bcolors
from stoobly_agent.lib.utils.conditional_decorator import ConditionalDecorator

from .helpers.endpoint_facade import EndpointFacade
from .helpers.endpoints_apply_context import EndpointsApplyContext
from .helpers.endpoints_import_context import EndpointsImportContext
from .helpers.feature_flags import local, remote
from .helpers.validations import validate_project_key, validate_scenario_key

settings = Settings.instance()
is_remote = remote(settings)
is_local = local(settings)

@click.group(
  epilog="Run 'stoobly-agent feature COMMAND --help' for more information on a command.",
  help="Manage endpoints"
)
@click.pass_context
def endpoint(ctx):
  pass

@endpoint.command(
  help="Apply endpoint to requests"
)
@click.option('--lifecycle-hooks-path', help='Path to lifecycle hooks script.')
@ConditionalDecorator(lambda f: click.option('--project-key', help='Project to create endpoint to.')(f), is_remote)
@ConditionalDecorator(lambda f: click.option('--remote-project-key', help='Which remote project to apply endpoints from.')(f), is_remote and is_local) 
@click.option('--scenario-key', help='Which scenario requests to apply the endpoint to. If none, then the endpoint will be applied to all requests.')
@click.option('--source-format', required=True, type=click.Choice([OPENAPI_FORMAT]), help='Spec file format.')
@click.option('--source-path', help='Path to spec file.')
def apply(**kwargs):
  context = EndpointsApplyContext()

  project_key = kwargs.get('remote_project_key') or settings.remote.project_key
  if project_key:
    project_key = validate_project_key(project_key)
    context.with_project(project_key.id)

  if kwargs.get('scenario_key'):
    scenario_key = validate_scenario_key(kwargs['scenario_key'])
    context.with_scenario(scenario_key.id)

  context.with_lifecycle_hooks_path(kwargs.get('lifecycle_hooks_path'))
  context.with_source(kwargs.get('source_path'), kwargs.get('source_format'))

  endpoint_handler = lambda endpoint: print(f"{bcolors.OKBLUE}Applying Endpoint {endpoint['method']} {endpoint['path']}{bcolors.ENDC}")
  context.with_endpoint_handler(endpoint_handler)

  request_handler = lambda request, endpoint: print(request.url)
  context.with_request_handler(request_handler)

  facade = EndpointFacade(settings)
  facade.apply(context)

@endpoint.command(
  "import",
  help="Import endpoints"
)
@click.option('--source-format', required=True, type=click.Choice([OPENAPI_FORMAT]), help="Spec file format.")
@click.option('--source-path', help='Path to spec file.')
def import_endpoint(**kwargs):
  context = EndpointsImportContext()

  project_key = settings.proxy.intercept.project_key
  if project_key:
    project_key = validate_project_key(project_key)
    context.with_project(project_key.id)
  
  context.with_source(kwargs.get('source_path'), kwargs.get('source_format'))

  endpoint_handler = lambda endpoint: print(f"{bcolors.OKBLUE}Importing Endpoint {endpoint['method']} {endpoint['path']}{bcolors.ENDC}")
  context.with_endpoint_handler(endpoint_handler)
  
  facade = EndpointFacade(settings)
  facade.import_endpoints(context)


import json
import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.models.adapters.orm import JoinedRequestStringAdapter
from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.scenario_snapshot import ScenarioSnapshot
from stoobly_agent.cli import record, scenario
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestRequest():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
      return reset()

    class TestWhenReplacing():

      @pytest.fixture(scope='class')
      def created_scenario(self, runner: CliRunner):
          create_result = runner.invoke(scenario, ['create', 'test'])
          assert create_result.exit_code == 0
          return Scenario.last()

      @pytest.fixture(scope='class', autouse=True)
      def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0
          return Request.last()

      @pytest.fixture(scope='class', autouse=True)
      def recorded_request_two(self, runner: CliRunner, created_scenario: Scenario):
          record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), NON_DETERMINISTIC_GET_REQUEST_URL])
          assert record_result.exit_code == 0
          return Request.last()

      @pytest.fixture(scope='class', autouse=True)
      def snapshot_result(self, runner: CliRunner, created_scenario: Scenario):
        snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
        assert snapshot_result.exit_code == 0
        return snapshot_result

      def test_it_snapshots_metadata(self, created_scenario: Scenario):
        snapshot = ScenarioSnapshot(created_scenario.uuid)

        scenario_metadata = snapshot.metadata
        
        assert scenario_metadata['name'] == created_scenario.name
        assert scenario_metadata['description'] == created_scenario.description

      def test_it_snapshots(self, created_scenario: Scenario, recorded_request_one: Request, recorded_request_two: Request):
        snapshot = ScenarioSnapshot(created_scenario.uuid)

        raw_requests = snapshot.requests

        assert raw_requests[0] == JoinedRequestStringAdapter(recorded_request_one).adapt()
        assert raw_requests[1] == JoinedRequestStringAdapter(recorded_request_two).adapt()

      def test_it_logs(self):
        log = Log()

        events = log.raw_events
        assert len(events) == 1

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 1

      def test_it_deletes(self, runner: CliRunner, created_scenario: Scenario):
        snapshot_result = runner.invoke(scenario, ['snapshot', created_scenario.key()])
        assert snapshot_result.exit_code == 0

        log = Log()

        events = log.raw_events
        assert len(events) == 2

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 1


from .app import App

class Command():

  def __init__(self, app: App):
    self.__app = app
    self.__namespace = app.scaffold_namespace

  @property
  def app(self):
    return self.__app

  @property
  def namespace(self):
    return self.__namespace


import json
import pdb
import re

from mitmproxy.http import Request as MitmproxyRequest
from requests import Response
from typing import List, TypedDict, Union

from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.proxy.mock.custom_not_found_response_builder import CustomNotFoundResponseBuilder
from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.app.settings.match_rule import MatchRule
from stoobly_agent.config.constants import custom_headers, query_params as request_query_params
from stoobly_agent.lib.api.param_builder import ParamBuilder
from stoobly_agent.lib.api.interfaces.requests import RequestResponseShowQueryParams
from stoobly_agent.lib.api.keys.project_key import InvalidProjectKey
from stoobly_agent.lib.api.keys.scenario_key import InvalidScenarioKey
from stoobly_agent.lib.logger import bcolors, Logger

from .hashed_request_decorator import HashedRequestDecorator
from .search_endpoint import inject_search_endpoint
from ..mitmproxy.request_facade import MitmproxyRequestFacade

LOG_ID = 'EvalRequest'

class EvalRequestOptions(TypedDict):
    infer: bool
    retry: int

def inject_eval_request(
    request_model: RequestModel,
    intercept_settings: InterceptSettings,
):
    settings = Settings.instance()

    if not request_model:
        request_model = RequestModel(settings)

    if not intercept_settings:
        intercept_settings = InterceptSettings(intercept_settings)

    return lambda request, ignored_components, **options: eval_request(
        request_model, intercept_settings, request, ignored_components or [], **options 
    )

def eval_request(
    request_model: RequestModel,
    intercept_settings: InterceptSettings,
    request: MitmproxyRequest,
    ignored_components_list: List[Union[list, str, None]],
    **options: EvalRequestOptions
) -> Response:
    query_params_builder = ParamBuilder({})
    scenario_key = intercept_settings.scenario_key

    if not scenario_key:
        Logger.instance(LOG_ID).info(f"{bcolors.WARNING}Missing{bcolors.ENDC} scenario key, defaulting to all requests for mocking")

    try:
        query_params_builder.with_resource_scoping(intercept_settings.project_key, scenario_key)
    except InvalidScenarioKey:
        Logger.instance(LOG_ID).warn(f"{bcolors.WARNING}Invalid{bcolors.ENDC} scenario key ${scenario_key}")
        # If project_key or scenario_key are invalid, assume custom not found
        return CustomNotFoundResponseBuilder().build()
    except InvalidProjectKey:
        Logger.instance(LOG_ID).warn(f"{bcolors.WARNING}Invalid{bcolors.ENDC} project key ${intercept_settings.project_key}")
        # If project_key or scenario_key are invalid, assume custom not found
        return CustomNotFoundResponseBuilder().build()

    # Tease out API returning ignored components on custom not found
    if request_model.is_local and not options.get('retry'):
        remote_project_key = intercept_settings.parsed_remote_project_key

        if remote_project_key:
            search_endpoint = inject_search_endpoint(intercept_settings)
            remote_project_id = remote_project_key.id
            endpoint_promise = lambda: search_endpoint(remote_project_id, request.method, request.url, ignored_components=1) 

            query_params_builder.with_param(request_query_params.ENDPOINT_PROMISE, endpoint_promise)

    ignored_components = __build_ignored_components(ignored_components_list or [])
    query_params_builder.with_params(__build_request_params(request, ignored_components))
    query_params_builder.with_params(__build_optional_params(request, options))

    query_params = query_params_builder.build()
    __filter_by_match_rules(request, intercept_settings.match_rules, query_params)

    return request_model.response(**query_params)

def __build_ignored_components(ignored_components_list):
    ignored_components = []
    for el in ignored_components_list:
        if isinstance(el, bytes) or isinstance(el, str): 
            try:
                ignored_components += json.loads(el)
            except:
                pass
        elif isinstance(el, list):
            ignored_components += el
        elif isinstance(el, dict):
            ignored_components.append(el)

    return ignored_components

###
#
# Formats request into parameters expected by stoobly request_model
#
# @param request [lib.mitmproxy_request_adapter.MitmproxyRequestFacade]
# @param ignored_components [Array<Hash>]
#
# @return [Hash] query parameters to pass to stoobly request_model
#
def __build_request_params(request: MitmproxyRequest, ignored_components = []) -> RequestResponseShowQueryParams:
    request = MitmproxyRequestFacade(request)
    hashed_request = HashedRequestDecorator(request).with_ignored_components(ignored_components)

    query_params = {}
    query_params['host'] = request.host
    query_params['path'] = request.path
    query_params['port'] = request.port
    query_params['method'] = request.method

    headers_hash = hashed_request.headers_hash()
    query_params['headers_hash'] = headers_hash

    query_params_hash = hashed_request.query_params_hash()
    query_params['query_params_hash'] = query_params_hash

    body_params_hash = hashed_request.body_params_hash()
    body_text_hash = hashed_request.body_text_hash()

    # If a body_param_hash is not set, then body_text_hash is checked
    # 
    # However, if body_params are ignored, then there may not be a body_param_hash
    # If there are ignored body_params, this means there are body_params
    # Then we should use body_params_hash over body_text_hash
    if len(body_params_hash) > 0 or len(hashed_request.ignored_body_params) > 0:
        query_params['body_params_hash'] = body_params_hash
    else:
        query_params['body_text_hash'] = body_text_hash

    return query_params

def __build_optional_params(request: MitmproxyRequest, options: EvalRequestOptions):
    optional_params = {}

    if options.get('retry'):
        optional_params['retry'] = options['retry']

        if options.get('infer'):
            optional_params['infer'] = options['infer']

    headers = request.headers
    if custom_headers.MOCK_REQUEST_ID in headers:
        optional_params['request_id'] = headers[custom_headers.MOCK_REQUEST_ID]

    if custom_headers.SESSION_ID in headers:
        optional_params[request_query_params.SESSION_ID] = headers[custom_headers.SESSION_ID]

    return optional_params

def __filter_by_match_rules(request: MitmproxyRequest, match_rules: List[MatchRule], query_params: RequestResponseShowQueryParams):
    components = [request_component.BODY_PARAM, request_component.HEADER, request_component.QUERY_PARAM]
    method = request.method.upper()

    keep = {
        request_component.BODY_PARAM: False,
        request_component.HEADER: False,
        request_component.QUERY_PARAM: False,
    }

    match_rules = list(filter(lambda rule: method in rule.methods, match_rules))
    for match_rule in match_rules:
        pattern = match_rule.pattern

        if not __matches(request.url, pattern):
            continue

        if not isinstance(match_rule.components, list):
            continue
 
        for component in components:
            keep[component] = component in match_rule.components

    if not keep[request_component.HEADER]:
        if 'headers_hash' in query_params:
            del query_params['headers_hash']

    if not keep[request_component.QUERY_PARAM]:
        if 'query_params_hash' in query_params:
            del query_params['query_params_hash']

    if not keep[request_component.BODY_PARAM]:
        if 'body_params_hash' in query_params:
            del query_params['body_params_hash']

        if 'body_text_hash' in query_params:
            del query_params['body_text_hash']

def __matches(url: str, pattern: str):
    try:
        return re.match(pattern, url)
    except re.error as e:
        Logger.instance().error(f"RegExp error '{e}' for {pattern}")
        return False


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.cli.helpers.handle_replay_service import BODY_FORMAT
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.cli import record, request, scenario
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.utils.decode import decode

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestReplay():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    class TestWhenNoOptions():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def recorded_response_raw(self, recorded_request: Request):
            return recorded_request.response.raw
            
        def test_it_matches_recorded_response(self, runner, recorded_request: Request, recorded_response_raw: bytes):
            replay_result = runner.invoke(request, ['replay', '--format', BODY_FORMAT, recorded_request.key()])
            assert replay_result.exit_code == 0

            response = RawHttpResponseAdapter(recorded_response_raw).to_response()
            assert replay_result.stdout.rstrip() == decode(response.content)

        def test_it_does_not_override_response_raw(self, recorded_request: Request, recorded_response_raw: bytes):
            new_response = Request.find(recorded_request.id).response.raw
            assert new_response == recorded_response_raw

    class TestWhenOverwritting():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def recorded_response_raw(self, recorded_request: Request):
            return recorded_request.response.raw

        def test_it_overwrites_response_raw(self, runner: CliRunner, recorded_request: Request, recorded_response_raw: bytes):
            replay_result = runner.invoke(request, ['replay', '--overwrite', recorded_request.key()])
            assert replay_result.exit_code == 0

            new_response = Request.find(recorded_request.id).response.raw
            assert new_response != recorded_response_raw

        def test_it_creates_one_replayed_response(self, recorded_request: Request):
            assert len(recorded_request.replayed_responses) == 1

        def test_it_sets_replayed_response_raw_to_response_raw(self, recorded_request: Request, recorded_response_raw: bytes):
            replayed_response = recorded_request.replayed_responses[0]
            assert replayed_response.raw == recorded_response_raw

    class TestWhenRecord():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_records(self, runner: CliRunner, recorded_request: Request):
            replay_result = runner.invoke(request, ['replay', '--record', recorded_request.key()])
            assert replay_result.exit_code == 0

            assert Request.count() == 2

        class TestWhenScenario():
            @pytest.fixture(scope='class', autouse=True)
            def settings(self):
                return reset()

            @pytest.fixture(scope='class')
            def recorded_request(self, runner: CliRunner):
                record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
                assert record_result.exit_code == 0
                return Request.last()

            @pytest.fixture()
            def scenario_key(self, runner: CliRunner):
                res = runner.invoke(scenario, ['create', '--select', 'key', '--without-headers', 'test-scenario'])
                assert res.exit_code == 0
                return ScenarioKey(res.stdout.strip())

            def test_it_records_to_scenario(self, runner: CliRunner, recorded_request: Request, scenario_key: ScenarioKey):
                replay_result = runner.invoke(request, ['replay', '--record' , '--scenario-key', scenario_key.raw, recorded_request.key()])
                assert replay_result.exit_code == 0

                assert Request.count() == 2

                _request = Request.last()
                assert _request.scenario.uuid == scenario_key.id

    class TestWhenScheme():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_overrides_scheme(self, runner: CliRunner, recorded_request: Request):
            replay_result = runner.invoke(request, ['replay', '--record', '--scheme' , 'http', recorded_request.key()])
            assert replay_result.exit_code == 0

            _request = Request.last()
            assert _request.scheme == 'http'

    class TestWhenHost():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_overrides_scheme(self, runner: CliRunner, recorded_request: Request):
            host = 'www.google.com'
            replay_result = runner.invoke(request, ['replay', '--record', '--host' , host, recorded_request.key()])
            assert replay_result.exit_code == 0

            _request = Request.last()
            assert _request.host == host


import os
import shutil

import pytest

from stoobly_agent.config.constants.env_vars import ENV
from stoobly_agent.config.constants.mode import NONE
from stoobly_agent.config.data_dir import DataDir, DATA_DIR_NAME
from stoobly_agent.test.test_helper import reset

class TestDataDir():
  @pytest.fixture(autouse=True, scope='class')
  def settings(self):
    return reset()

  @pytest.fixture(scope='class')
  def original_cwd(self) -> str:
    return os.getcwd()

  @pytest.fixture(scope='class')
  def home_dir(self) -> str:
    return os.path.expanduser("~")

  def test_in_home(self, original_cwd: str, home_dir: str):
    # A previous test can put us in 'stoobly_agent/test/app/models/schemas/.stoobly'
    os.chdir(original_cwd)
    data_dir_path = os.path.join(home_dir, DATA_DIR_NAME)
    os.environ[ENV] = NONE

    result = DataDir.instance().path

    assert result == data_dir_path

  def test_in_cwd(self, original_cwd: str):
    os.environ[ENV] = NONE
    DataDir._instances = None

    temp_dir = os.path.join(original_cwd, 'tmp')
    nested_temp_dir = os.path.join(temp_dir, 'tmp-nested')
    data_dir_path = os.path.join(nested_temp_dir, DATA_DIR_NAME)
    os.makedirs(data_dir_path)

    # Go into nested folder
    os.chdir(nested_temp_dir)

    try:
      result = DataDir.instance().path

      assert result == data_dir_path
      assert result != temp_dir

    finally:
      shutil.rmtree(temp_dir)
      os.chdir(original_cwd)

  def test_in_parent_nested(self, original_cwd: str):
    os.environ[ENV] = NONE
    DataDir._instances = None

    # Create a temporary directory structure for testing
    temp_dir = os.path.join(original_cwd, 'tmp')

    # Create nested folders
    nested_dir1 = os.path.join(temp_dir, "nested1")
    nested_dir2 = os.path.join(nested_dir1, "nested2")
    nested_dir3 = os.path.join(nested_dir2, "nested3")
    os.makedirs(nested_dir3)

    # Create .stoobly folder in the nested structure
    data_dir_path = os.path.join(nested_dir1, DATA_DIR_NAME)
    os.makedirs(data_dir_path)

    # Go into dir3
    os.chdir(nested_dir3)

    try:
      # Test when .stoobly folder exists in the nested structure
      result = DataDir.instance().path

      assert result == data_dir_path
      assert result != temp_dir

    finally:
      shutil.rmtree(temp_dir)
      os.chdir(original_cwd)


from typing import Union

from stoobly_agent.lib.utils.decode import decode

class ResponseStringControl():
  RESPONSE_TYPE = 2

  def __init__(self, c: Union[bytes, str] = None):
    self.__latency = 0 # Millisecond
    self.__response_type = self.RESPONSE_TYPE
    self.__id = None
    self.__timestamp = 0 # Microsecond

    if c:
      self.parse(c)

  @property
  def id(self):
    return self.__id

  @id.setter
  def id(self, id: str):
    self.__id = id

  @property
  def latency(self):
    return self.__latency

  @latency.setter
  def latency(self, l: int):
    self.__latency = l

  @property
  def response_type(self):
    return self.__response_type

  @property
  def timestamp(self):
    return self.__timestamp

  @timestamp.setter
  def timestamp(self, t: int):
    self.__timestamp = t

  def parse(self, s: str):
    toks = decode(s).split(' ')
    self.__id = toks[1]
    self.__timestamp = int(toks[2])

    if len(toks) > 3:
      self.__latency = int(toks[3])

  def serialize(self):
    return "{} {} {} {}".format(self.RESPONSE_TYPE, self.__id, self.__timestamp, self.__latency)


import pdb

from typing import Tuple

from stoobly_agent.app.models.types import ScenarioDestroyParams, ScenarioCreateParams
from stoobly_agent.lib.api.interfaces import ScenariosIndexQueryParams, ScenariosIndexResponse, ScenarioShowResponse
from stoobly_agent.lib.orm import ORM
from stoobly_agent.lib.orm.scenario import Scenario

from .helpers.search import search_scenario
from .helpers.snapshot_service import snapshot_scenario
from .local_db_adapter import LocalDBAdapter

class LocalDBScenarioAdapter(LocalDBAdapter):
  __scenario_orm = None

  def __init__(self, scenario_orm: Scenario.__class__ = Scenario):
    self.__scenario_orm = scenario_orm

  def create(self, **params: ScenarioCreateParams) -> Tuple[ScenarioShowResponse, int]:
    with ORM.instance().db.transaction():
      scenario_record = self.__scenario_orm.create(params)
      return self.success(scenario_record.to_dict())

  def show(self, scenario_id: str) -> Tuple[ScenarioShowResponse, int]:
    scenario_record = self.__scenario(scenario_id)
    if not scenario_record:
      return self.__scenario_not_found()
    return self.success(self.__to_show_response(scenario_record))

  def index(self, **query_params: ScenariosIndexQueryParams) -> Tuple[ScenariosIndexResponse, int]:
    page = int(query_params.get('page') or 0)
    query = query_params.get('q')
    size = int(query_params.get('size') or 20)
    sort_by = query_params.get('sort_by') or 'id'
    sort_order = query_params.get('sort_order') or 'desc'

    filter = query_params.get('filter')
    is_deleted = filter == 'is_deleted'

    scenarios = Scenario.where('is_deleted', is_deleted)

    if filter == 'starred':
      scenarios = scenarios.where('starred', True)
    elif filter == 'high_priority':
      scenarios = scenarios.where('priority', 3)
    elif filter == 'medium_priority':
      scenarios = scenarios.where('priority', 2)
    elif filter == 'low_priority':
      scenarios = scenarios.where('priority', 1)
    elif filter == 'none_priority':
      scenarios = scenarios.where('priority', 0)

    if query:
      scenarios = search_scenario(scenarios, query)

    total = scenarios.count()
    scenarios = scenarios.offset(page * size).limit(size).order_by(sort_by, sort_order).get()

    return self.success({
      'list': list(map(lambda scenario: self.__to_show_response(scenario), scenarios.items)),
      'total': total,
    })

  def update(self, scenario_id: int, **params: ScenarioCreateParams) -> Tuple[ScenarioShowResponse, int]:
    scenario = self.__scenario(scenario_id)

    if not scenario:
      return self.__scenario_not_found()

    if scenario.update(params):
      return self.success(self.__to_show_response(scenario))

    return self.internal_error('Could not update scenario')

  def destroy(self, scenario_id: int, **params: ScenarioDestroyParams) -> Tuple[ScenarioShowResponse, int]:
    scenario = self.__scenario(scenario_id)

    if not scenario:
      return self.__scenario_not_found()

    if params.get('force') or scenario.is_deleted:
      scenario.delete()
    else:
      scenario.update({'is_deleted': True})

    return self.success(self.__to_show_response(scenario))

  def snapshot(self, scenario_id: str, **params):
    scenario = self.__scenario(scenario_id)

    if not scenario:
      return self.__scenario_not_found()

    file_path = snapshot_scenario(scenario, **params)
    if not file_path:
      return self.internal_error()

    return self.success(file_path)

  def __scenario(self, scenario_id: str):
    if self.validate_uuid(scenario_id):
      return self.__scenario_orm.find_by(uuid=scenario_id)
    else:
      return self.__scenario_orm.find(scenario_id)

  def __to_show_response(self, scenario: Scenario) -> ScenarioShowResponse:
    res = scenario.to_dict()
    return res

  def __scenario_not_found(self):
    return self.not_found('Scenario not found')


import pdb
import requests

from typing import Union

from stoobly_agent.app.settings import Settings

from .types import ScenarioDestroyParams, ScenarioCreateParams
from .factories.resource.scenario import ScenarioResourceFactory
from .model import Model

class ScenarioModel(Model):

  def __init__(self, settings: Settings, **options):
    super().__init__(settings, **options)

  def as_local(self):
    self.adapter = ScenarioResourceFactory(self.settings.remote).local_db()
    self.is_local = True

  def as_remote(self):
    self.adapter = ScenarioResourceFactory(self.settings.remote).stoobly()
    self.is_local = False

  def create(self, **body_params: ScenarioCreateParams):
    try:
      return self.adapter.create(**body_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def show(self, scenario_id: str):
    try:
      return self.adapter.show(scenario_id)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def index(self, **query_params):
    try:
      return self.adapter.index(**query_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def update(self, scenario_id: str, **params: ScenarioCreateParams):
    try:
      return self.adapter.update(scenario_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def destroy(self, scenario_id, **params: ScenarioDestroyParams):
    try:
      return self.adapter.destroy(scenario_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def snapshot(self, scenario_id: str, **params):
    try:
      return self.adapter.snapshot(scenario_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)   


import pdb

from mitmproxy.http import Response as MitmproxyResponse
from typing import Union

from ..replay.body_parser_service import decode_response, encode_response

class MitmproxyResponseBodyFacade:
    def __init__(self, response: MitmproxyResponse):
        self.__response = response

    def get(self, content_type: Union[bytes, str]):
        return decode_response(self.__response.content, content_type)

    def set(self, content, content_type: Union[bytes, str]):
        """
        Adjusting Content-Length header should be done by MitmproxyResponse
        """
        self.__response.content = encode_response(content, content_type).encode()


import pdb

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from mitmproxy.http import Response as MitmproxyResponse
from requests import Response

###
#
# Return response headers, body, and status code
#
def pass_on(flow: MitmproxyHTTPFlow, res: Response):
    if not isinstance(res, Response):
        return
  
    headers = {}
    for key, value in res.headers.items():
        headers[key.title()] = value

    # Without specifying a length to read, requests will compare content length
    # with Content-Length header. If the content is gzipped, an IncompleteRead error will be thrown
    #content = res.raw.read(res.raw.length_remaining)

    # Ideally we just return the HTTPResponse object from res.raw
    # See Issue #11
    # content = res.raw
    content = res.content

    flow.response = MitmproxyResponse.make(
        res.status_code, content, headers,
    )

def bad_request(flow: MitmproxyHTTPFlow, message: str):
    flow.response = MitmproxyResponse.make(
        400,  # (optional) status code
        message,
        {'Content-Type': 'text/plain'}  # (optional) headers
    )

def enable_cors(flow: MitmproxyHTTPFlow):
    flow.response = MitmproxyResponse.make(
        200,
        '',
        {
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': 'GET, OPTIONS, POST, PATCH, PUT, DELETE',
            'Access-Control-Allow-Headers': '*'
        }
    )

# Without deleting this header, causes parsing issues when reading response
def disable_transfer_encoding(response: MitmproxyResponse) -> None:
    header_name = 'Transfer-Encoding'
    if header_name in response.headers and response.headers[header_name] == 'chunked':
        del response.headers['Transfer-Encoding']


import pytest

from stoobly_agent.app.cli.helpers.schema_builder import SchemaBuilder

@pytest.mark.openapi
class TestSchemaBuilder():

  def test_it_builds_single_level_dict(self):
    builder = SchemaBuilder(-1, 'query_param_name')
    param_names = builder.build([
      {'name': 'id', 'value': 0, 'query': 'id', 'required': False, 'id': 1, "parent_id": None},
      {'name': 'name', 'value': '', 'query': 'name', 'required': False, 'id': 2, "parent_id": None}
    ])

    param_names_index = self.__index(param_names)
    expected_queries = ['id.Integer', 'name.String']

    for query in expected_queries:
      assert param_names_index.get(query), query

  def test_it_builds_two_level_dict(self):
    builder = SchemaBuilder(-1, 'query_param_name')
    param_names = builder.build([
      {'name': 'list', 'value': [], 'query': 'list', 'required': False, 'id': 1, "parent_id": None},
      {'name': 'ListElement', 'value': {}, 'query': 'list[*]', 'required': False, 'id': 2, "parent_id": 1},
      {'name': 'id', 'value': 0, 'query': 'list[*].id', 'required': False, 'id': 3, "parent_id": 2},
      {'name': 'name', 'value': '', 'query': 'list[*].name', 'required': False, 'id': 4, "parent_id": 2},
      {'name': 'total', 'value': 0, 'query': 'total', 'required': False, 'id': 5, "parent_id": None}
    ])

    param_names_index = self.__index(param_names)
    expected_queries = [
      'list.Array', 'list[*].Hash', 'list[*].id.Integer', 'list[*].name.String', 'total.Integer'
    ]

    for query in expected_queries:
      assert param_names_index.get(query), query

  def test_it_builds_single_level_array(self):
    builder = SchemaBuilder(-1, 'body_param_name')
    param_names = builder.build([
      {'name': 'Element', 'value': 0, 'query': '[*]', 'required': False, 'id': 1, "parent_id": None},
    ])

    param_names_index = self.__index(param_names)
    expected_queries = [
      '[*].Integer'
    ]

    for query in expected_queries:
      assert param_names_index.get(query), query

  def test_it_builds_item_and_single_level_array(self):
    builder = SchemaBuilder(-1, 'query_param_name')
    param_names = builder.build([
      {'name': 'tags', 'value': [], 'query': 'tags', 'required': False, 'id': 1, "parent_id": None},
      {'name': 'TagsElement', 'value': '', 'query': 'tags[*]', 'required': False, 'id': 2, "parent_id": 1},
      {'name': 'limit', 'value': 0, 'query': 'limit', 'required': False, 'id': 3, "parent_id": None}
    ])

    param_names_index = self.__index(param_names)
    expected_queries = [
      'tags.Array', 'tags[*].String', 'limit.Integer' 
    ]

    for query in expected_queries:
      assert param_names_index.get(query), query

  def __index(self, param_names):
    param_names_index = {}
    for param_name in param_names:
      param_names_index[f"{param_name['query']}.{param_name['inferred_type']}"] = param_name
    return param_names_index



import json
import pdb
import pytest

from mitmproxy.http import Headers, Request as MitmproxyRequest
from time import time

from stoobly_agent.app.settings.constants import intercept_mode, request_component
from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.settings.rewrite_rule import RewriteRule

@pytest.fixture
def mitmproxy_get_request():
  now = time()
  return MitmproxyRequest(
    'stoobly.com',
    443,
    b'GET',
    b'https',
    b'stoobly.com:443',
    b'/requests?project_id=1',
    b'HTTP/1.1',
    Headers(**{'access-token': b'123'}),
    b'',
    Headers(), # Trailers
    now,
    now + 1,
  )

@pytest.fixture
def mitmproxy_post_request():
  now = time()
  return MitmproxyRequest(
    'stoobly.com',
    443,
    b'POST',
    b'https',
    b'stoobly.com:443',
    b'/users?project_id=1',
    b'HTTP/1.1',
    Headers(**{'content-type': b'application/json'}),
    json.dumps({'name': 'Alice', 'age': 20}).encode(),
    Headers(), # Trailers
    now,
    now + 1,
  )

class TestRewriteParams():

  def test_rewrites_header(self, mitmproxy_get_request: MitmproxyRequest):
    parameter_rule = {
      'modes': [intercept_mode.RECORD],
      'name': 'access-token',
      'type': request_component.HEADER,
      'value': '',
    }
    rewrite_rule = RewriteRule({
      'methods': ['GET'],
      'pattern': '.*?/requests',
      'parameter_rules': [parameter_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_get_request)
    facade.with_parameter_rules([rewrite_rule])
    facade.rewrite()

    headers = facade.headers
    assert headers['access-token'] == ''

  def test_rewrites_query_param(self, mitmproxy_get_request: MitmproxyRequest):
    parameter_rule = {
      'modes': [intercept_mode.RECORD],
      'name': 'project_id',
      'type': request_component.QUERY_PARAM,
      'value': '',
    }
    rewrite_rule = RewriteRule({
      'methods': ['GET'],
      'pattern': '.*?/requests',
      'parameter_rules': [parameter_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_get_request)
    facade.with_parameter_rules([rewrite_rule])
    facade.rewrite()

    query_params = facade.query
    assert query_params['project_id'] == ''
 
  def test_rewrites_body_param(self, mitmproxy_post_request: MitmproxyRequest):
    parameter_rule = {
      'modes': [intercept_mode.RECORD],
      'name': 'age',
      'type': request_component.BODY_PARAM,
      'value': '',
    }
    rewrite_rule = RewriteRule({
      'methods': ['POST'],
      'pattern': '.*?/users',
      'parameter_rules': [parameter_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_post_request)
    facade.with_parameter_rules([rewrite_rule])
    facade.rewrite()

    body = facade.body
    body_params = json.loads(body)
    assert body_params['age'] == '' 

class TestRewriteUrl():

  def test_rewrites_hostname(self, mitmproxy_get_request: MitmproxyRequest):
    hostname = 'test'
    url_rule = {
      'modes': [intercept_mode.RECORD],
      'hostname': hostname,
    }
    rewrite_rule = RewriteRule({
      'methods': ['GET'],
      'pattern': '.*?/requests',
      'url_rules': [url_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_get_request)
    facade.with_url_rules([rewrite_rule])

    assert facade.host != hostname

    facade.rewrite()

    assert facade.host == hostname

  def test_rewrites_port(self, mitmproxy_get_request: MitmproxyRequest):
    port = '8000'
    url_rule = {
      'modes': [intercept_mode.RECORD],
      'port': port,
    }
    rewrite_rule = RewriteRule({
      'methods': ['GET'],
      'pattern': '.*?/requests',
      'url_rules': [url_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_get_request)
    facade.with_url_rules([rewrite_rule])

    assert facade.port != int(port)

    url_before = facade.url
    facade.rewrite()

    assert facade.port == int(port)
    assert url_before != facade.url

  def test_rewrites_scheme(self, mitmproxy_get_request: MitmproxyRequest):
    scheme = 'http'
    url_rule = {
      'modes': [intercept_mode.RECORD],
      'scheme': scheme,
    }
    rewrite_rule = RewriteRule({
      'methods': ['GET'],
      'pattern': '.*?/requests',
      'url_rules': [url_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_get_request)
    facade.with_url_rules([rewrite_rule])

    assert facade.scheme != scheme

    url_before = facade.url
    facade.rewrite()

    assert facade.scheme == scheme
    assert url_before != facade.url

  def test_rewrites_path(self, mitmproxy_get_request: MitmproxyRequest):
    path = '/index.html'
    url_rule = {
      'modes': [intercept_mode.RECORD],
      'path': path
    }
    rewrite_rule = RewriteRule({
      'methods': ['GET'],
      'pattern': '.*?/requests',
      'url_rules': [url_rule]
    })

    facade = MitmproxyRequestFacade(mitmproxy_get_request)
    facade.with_url_rules([rewrite_rule])

    assert facade.path != path

    url_before = facade.url
    facade.rewrite()

    assert facade.path == path
    assert url_before != facade.url

  class TestWhenMultipleRules():

    def test_rewrites_hostname(self, mitmproxy_get_request: MitmproxyRequest):
      hostname1 = 'test1'
      url_rule1 = {
        'modes': [intercept_mode.RECORD],
        'hostname': hostname1,
      }
      hostname2 = 'test2'
      url_rule2 = {
        'modes': [intercept_mode.RECORD],
        'hostname': hostname2,
      }
      rewrite_rule = RewriteRule({
        'methods': ['GET'],
        'pattern': '.*?/requests',
        'url_rules': [url_rule1, url_rule2]
      })

      facade = MitmproxyRequestFacade(mitmproxy_get_request)
      facade.with_url_rules([rewrite_rule])
      facade.rewrite()

      assert facade.host == hostname2


import mimetypes
import os
import pdb
import re

from io import BytesIO
from mitmproxy.http import Request as MitmproxyRequest
from requests import Response
from requests.structures import CaseInsensitiveDict
from typing import Union

from stoobly_agent.lib.logger import bcolors, Logger

from .types import Fixtures

LOG_ID = 'Fixture'

class Options():
  public_directory_path: str
  response_fixtures: Fixtures

def eval_fixtures(request: MitmproxyRequest, **options: Options) -> Union[Response, None]:
  fixture_path = None
  headers = CaseInsensitiveDict()
  status_code = 200

  response_fixtures = options.get('response_fixtures')
  fixture: dict = __eval_response_fixtures(request, response_fixtures)

  if not fixture:
    public_directory_path = options.get('public_directory_path')

    if not public_directory_path:
      return

    request_path = 'index' if request.path == '/' else request.path
    _fixture_path = os.path.join(public_directory_path, request_path.lstrip('/'))
    if request.headers.get('accept'):
      fixture_path = __guess_file_path(_fixture_path, request.headers['accept'])

    if not fixture_path:
      fixture_path = _fixture_path

    if not os.path.isfile(fixture_path):
      return
  else:
    fixture_path = fixture.get('path')
    if not fixture_path or not os.path.isfile(fixture_path):
      return

    _headers = fixture.get('headers')
    headers = CaseInsensitiveDict(_headers if isinstance(_headers, dict) else {}) 

    if fixture.get('status_code'):
      status_code = fixture.get('status_code')

  with open(fixture_path, 'rb') as fp:
    response = Response()

    response.status_code = status_code
    response.raw = BytesIO(fp.read()) 
    response.headers = headers

    if not response.headers.get('content-type'):
      content_type = __guess_content_type(fixture_path)
      if content_type:
        response.headers['content-type'] = content_type

    Logger.instance(LOG_ID).debug(f"{bcolors.OKBLUE}Resolved{bcolors.ENDC} fixture {fixture_path}")

    return response

def __guess_content_type(file_path):
  file_extension = os.path.splitext(file_path)[1]
  if not file_extension:
    return
  return mimetypes.types_map.get(file_extension)

def __guess_file_path(file_path, content_type):
  file_extension = os.path.splitext(file_path)[1]
  if file_extension:
    return file_path

  if not content_type:
    return

  content_types = __parse_accept_header(content_type)

  for content_type in content_types:
    file_extension = mimetypes.guess_extension(content_type)

    if not file_extension:
      continue

    _file_path = f"{file_path}{file_extension}"
    if os.path.isfile(_file_path):
      return _file_path

def __eval_response_fixtures(request: MitmproxyRequest, response_fixtures: Fixtures):
  if not isinstance(response_fixtures, dict):
    return

  method = request.method
  routes = response_fixtures.get(method)

  if not isinstance(routes, dict):
    return

  for path_pattern in routes:
    if not re.match(path_pattern, request.path):
      continue
      
    fixture = routes[path_pattern]
    if not isinstance(fixture, dict):
      continue

    path = fixture.get('path')

    if path:
      return fixture

def __parse_accept_header(accept_header):
    types = []
    for item in accept_header.split(","):
        parts = item.split(";")
        content_type = parts[0].strip()
        q_value = 1.0  # Default quality value
        if len(parts) > 1 and parts[1].strip().startswith("q="):
            try:
                q_value = float(parts[1].strip()[2:])
            except ValueError:
                pass  # Keep default q_value if parsing fails
        types.append((content_type, q_value))

    # Sort by quality factor in descending order
    return [content_type for content_type, _ in sorted(types, key=lambda x: x[1], reverse=True)]



from .config import Config
from .constants import ENV_FILE
from .env import Env

class WorkflowEnv(Config):

  def __init__(self, dir: str):
    super().__init__(dir, ENV_FILE)

    self.__env_vars = {}

  def get(self, key):
    return self.__env_vars.get(key)

  def load(self):
    self.__env_vars = self.read()

  def write(self, env_vars: dict):
    Env(self.path).write(env_vars)
    self.__env_vars = env_vars


import requests

from stoobly_agent.lib.api.interfaces import RequestShowResponse
from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory

class TestStooblyAdapter():

  def test_www_google_com(self):
    request = requests.Request(
      data=b'',
      headers={
        'Content-Length': '0'
      },
      method='GET',
      url='https://www.google.com',
    )

    mitmproxy_request = PythonRequestAdapterFactory(request).stoobly_request()
    self.__test_equivalence(mitmproxy_request, request)

  def test_stoobly_com(self):
    data = b'{"name": "test", "description": "test"}'
    request = requests.Request(
      data=data,
      headers={
          'Content-Length': str(len(data)),
          'Content-Type': 'application/json',
      },
      method='POST',
      url='https://stoobly.com/scenarios',
    )

    mitmproxy_request = PythonRequestAdapterFactory(request).stoobly_request()

    self.__test_equivalence(mitmproxy_request, request)

  def __test_equivalence(self, stoobly_request: RequestShowResponse, request: requests.Request):
    assert stoobly_request['method'] == request.method
    assert stoobly_request['url'] == request.url
    assert stoobly_request['body'] == request.data

    stoobly_request_headers = {}
    for header in stoobly_request['headers']:
      stoobly_request_headers[header['name']] = header['value']
      
    for name, value in request.headers.items():
      assert value == stoobly_request_headers.get(name)




import os
import pdb

from ...constants import (
  COMPOSE_TEMPLATE, SERVICE_HOSTNAME, 
  SERVICE_HOSTNAME_ENV, SERVICE_NAME_ENV, SERVICE_PORT, SERVICE_PORT_ENV, SERVICE_SCHEME, SERVICE_SCHEME_ENV, 
  WORKFLOW_CONTAINER_CONFIGURE_TEMPLATE, WORKFLOW_CONTAINER_INIT_TEMPLATE, WORKFLOW_CONTAINER_PROXY_TEMPLATE, WORKFLOW_NAME, WORKFLOW_NAME_ENV
)
from ..builder import Builder
from ..service.builder import ServiceBuilder

class WorkflowBuilder(Builder):

  def __init__(self, workflow_path: str, service_builder: ServiceBuilder):
    self._env = [SERVICE_NAME_ENV, WORKFLOW_NAME_ENV]
    self.__workflow_name = os.path.basename(workflow_path)
    super().__init__(workflow_path, COMPOSE_TEMPLATE.format(workflow=self.__workflow_name))

    self.__context = '../'
    self.__profiles = [WORKFLOW_NAME]

    if not service_builder:
      service_path = os.path.dirname(workflow_path)
      service_builder = ServiceBuilder(service_path)

    self.__service_builder = service_builder

  @property
  def app(self):
    return f"{self.namespace}.app"

  @property
  def base_compose_file_path(self):
    return os.path.relpath(self.service_builder.compose_file_path, self.dir_path)

  @property
  def config(self):
    return self.service_builder.config

  @property
  def configure(self):
    return WORKFLOW_CONTAINER_CONFIGURE_TEMPLATE.format(service_name=self.namespace)

  @property
  def context(self):
    return self.__context

  @property
  def context_docker_file_path(self):
    return os.path.relpath(self.service_builder.app_builder.context_docker_file_path, self.service_path)

  @property
  def init(self):
    return WORKFLOW_CONTAINER_INIT_TEMPLATE.format(service_name=self.namespace)

  @property
  def namespace(self):
    return f"{self.service_builder.service_name}"

  @property
  def profiles(self):
    return self.__profiles

  @property
  def proxy(self):
    return WORKFLOW_CONTAINER_PROXY_TEMPLATE.format(service_name=self.namespace)

  @property
  def service_builder(self):
    return self.__service_builder

  @property
  def service_path(self):
    return self.__service_builder.dir_path

  @property
  def workflow_name(self):
    return self.__workflow_name

  def build_all(self):
    # Resources
    if self.config.detached:
      self.with_volume(self.service_builder.service_name)

    # Services
    self.build_init()
    self.build_configure()

    if self.config.hostname:
      self.build_proxy() # Depends on configure, must call build_configure first

  def build_init(self):
    # If the init_base service does not exist, we can't extend from it, return
    if not self.service_builder.init_base_service:
      return

      return
    service = {
      'extends': self.service_builder.build_extends_init_base(self.dir_path),
      'profiles': self.profiles,
    }

    self.with_service(self.init, service)

  def build_configure(self):
    # If the configure_base service does not exist, we can't extend from it, return
    if not self.service_builder.configure_base_service:
      return

    depends_on = {}

    service = {
      'depends_on': depends_on,
      'extends': self.service_builder.build_extends_configure_base(self.dir_path),
      'profiles': self.profiles,
    }

    if self.init in self.services:
      depends_on[self.init] = {
        'condition': 'service_completed_successfully',
      }

    self.with_service(self.configure, service)

  def build_proxy(self):
    # If the proxy_base service does not exist, we can't extend from it, return
    if not self.service_builder.proxy_base_service:
      return

    depends_on = {}
    networks = {}

    service = {
      'depends_on': depends_on,
      'extends': self.service_builder.build_extends_proxy_base(self.dir_path),
      'networks': networks,
      'profiles': self.profiles,
    }

    if self.configure in self.services:
      depends_on[self.init] = {
        'condition': 'service_completed_successfully',
      }

      depends_on[self.configure] = {
        'condition': 'service_completed_successfully',
      }

    # Expose this container service to the public network 
    # so that it is accessible to other Stoobly services
    networks[self.egress_network_name] = {}

    self.with_service(self.proxy, service)

  def env_dict(self):
    env = {}
    for e in self._env:
      env[e] = '${' + e + '}'
    return env

  def initialize_custom_file(self):
    dest = self.custom_compose_file_path

    if not os.path.exists(dest):
      compose = {
        'services': {}
      }

      if self.networks:
        compose['networks'] = self.networks

      super().write(compose, dest)

  def write(self):
    compose = {
      'services': self.services,
    }

    if self.networks:
      compose['networks'] = self.networks

    if self.volumes:
      compose['volumes'] = self.volumes

    super().write(compose)

  def __with_url_environment(self, environment):
    environment[SERVICE_HOSTNAME_ENV] = SERVICE_HOSTNAME

    if self.config.scheme:
      environment[SERVICE_SCHEME_ENV] = SERVICE_SCHEME

    if self.config.port:
      environment[SERVICE_PORT_ENV] = SERVICE_PORT



import pdb

from tabulate import tabulate
from typing import List

from ..types.print_options import TabulatePrintOptions

def tabulate_print(records: List[dict], **kwargs: TabulatePrintOptions):
  show_header = kwargs.get('headers')
  filter = kwargs.get('filter') or []
  format = kwargs.get('format') or 'plain'
  select = kwargs.get('select') or []
  print_handler = kwargs.get('print_handler') or print

  headers = []
  rows = []
  for record in records:
    build_header = show_header and len(headers) == 0

    row = []
    for key, val in record.items():
      # If exclude list is provided, check if key is excluded
      if key in filter:
        continue
      
      # If include list is provided, check if key is included
      if len(select) > 0 and key not in select:
        continue

      if build_header:
        headers.append(key)

      row.append(val)

    rows.append(row)

  print_handler(tabulate(rows, headers=headers, tablefmt=format))


import pdb

from runpy import run_path
from typing import Callable

from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.app.models.types import OPENAPI_FORMAT
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.app.settings import Settings

from .openapi_endpoint_adapter import OpenApiEndpointAdapter

class EndpointsApplyContext:

  def __init__(self):
    self.endpoints = []
    self.endpoint_handlers  = []
    self.lifecycle_hooks = {}
    self.request_handlers = []
    self.request_model = None
    self.scenario_id = None

  def with_endpoint_handler(self, handler: Callable[[EndpointShowResponse], None]):
    self.endpoint_handlers.append(handler)
    return self

  def with_lifecycle_hooks_path(self, path: str):
    try:
      self.lifecycle_hooks = run_path(path)
    except Exception as e:
      pass
  
    return self

  def with_project(self, project_id: int):
    settings = Settings.instance()

    resource = EndpointsResource(settings.remote.api_url, settings.remote.api_key)
    res = resource.index(project_id=project_id, ignored_components=1)
    res.raise_for_status()  
    
    self.endpoints += res.json()

    return self

  def with_request_handler(self, handler: Callable[[Request, EndpointShowResponse], None]):
    self.request_handlers.append(handler)
    return self

  def with_request_model(self, model):
    self.request_model = model
    return self

  def with_scenario(self, scenario_id: int):
    self.scenario_id = scenario_id
    return self

  def with_source(self, path: str, format: str):
    if format == OPENAPI_FORMAT:
      endpoint_adapter = OpenApiEndpointAdapter()
      self.endpoints += endpoint_adapter.adapt_from_file(path)

    return self


import pdb
import requests

from typing import Tuple

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.lib.orm.request import Request

from .local_db_adapter import LocalDBAdapter 
from .request_adapter import LocalDBRequestAdapter

class LocalDBBodyAdapter(LocalDBAdapter):
  __request_orm = None

  def __init__(self, request_orm: Request.__class__ = Request):
    self.__request_orm = request_orm

  def update(self, request_id, text: str):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    return self.success(LocalDBRequestAdapter(self.__request_orm).update(request_id, body=text))

  def mock(self, request_id) -> Tuple[requests.Request, int]:
    request = self.__request_orm.find(request_id)

    if not request:
      return self.__request_not_found()

    return self.success(RawHttpRequestAdapter(request.raw).to_request())
  
  def __request_not_found(self):
    return self.not_found('Request not found')  


import os
import pdb

from stoobly_orator import DatabaseManager, Model

from stoobly_agent.config.constants.env_vars import ENV
from stoobly_agent.config.data_dir import DataDir

class ORM():
    _instance = None
    _model = None
    _db = None

    def __init__(self):
        if ORM._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.initialize_db()
            self._model = Model
            
    def initialize_db(self):
        config = {
            'default': os.environ.get(ENV) or 'production',
            'test': {
                'driver': 'sqlite',
                'database': DataDir.instance().db_file_path,
            },
            'production': {
                'driver': 'sqlite',
                'database': DataDir.instance().db_file_path,
            }
        }

        db = DatabaseManager(config)
        Model.set_connection_resolver(db)

        self._db = db

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    @property
    def base(self):
        return self._model

    @property
    def db(self):
        return self._db


import json
import pdb
import pytest
import requests

from click.testing import CliRunner

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.models.adapters.orm.request.python_adapter import PythonRequestAdapter
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.orm.request import Request

from stoobly_agent.app.models.factories.resource.local_db.helpers.request_builder import RequestBuilder

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestOrmPythonRequestAdapter():
  @pytest.fixture(scope='class')
  def settings(self):
    return reset()

  @pytest.fixture(scope='class')
  def request_method(self):
    return 'POST'

  @pytest.fixture(scope='class')
  def request_url(self):
    return 'https://petstore.swagger.io/v2/pets'

  @pytest.fixture(scope='class')
  def request_body(self):
    return {
      'first_name': 'John',
      'last_name': 'Smith',
    }

  @pytest.fixture(scope='class')
  def request_headers(self):
    return {
      'Content-Type': 'application/json'
    }

  @pytest.fixture(scope='class')
  def created_request(
    self, settings: Settings, 
    request_method: str, request_url: str, request_headers: dict, request_body: dict
  ):
    status = RequestBuilder(
      method=request_method,
      request_body=request_body,
      request_headers=request_headers,
      response_body='',
      status_code=200,
      url=request_url,
    ).with_settings(settings).build()[1]
    assert status == 200

    return Request.last()

  @pytest.fixture()
  def python_request(self, created_request: Request):
    return PythonRequestAdapter(created_request).adapt()

  def test_it_sets_method(self, python_request: requests.Request, request_method: str):
    assert python_request.method == request_method

  def test_it_sets_url(self, python_request: requests.Request, request_url: str):
    assert python_request.url == request_url

  def test_it_sets_headers(self, python_request: requests.Request, request_headers: dict):
    data = python_request.data
    headers = python_request.headers
    assert headers == {
      **request_headers,
      'Content-Length': str(len(data))
    }

  def test_it_returns_data_as_bytes(self, python_request: requests.Request):
    assert isinstance(python_request.data, bytes)

  def test_it_sets_data(self, python_request: requests.Request, request_body: dict):
    data = python_request.data
    assert json.dumps(request_body) == data.decode()


from stoobly_agent.app.settings import RemoteSettings
from stoobly_agent.lib.orm.replayed_response import ReplayedResponse

from .local_db.replayed_response_adapter import LocalDBReplayedResponseAdapter

class ReplayedResponseResourceFactory():

  def __init__(self, settings: RemoteSettings):
    self.__remote_settings = settings

  def local_db(self) -> LocalDBReplayedResponseAdapter:
    return LocalDBReplayedResponseAdapter(ReplayedResponse)  


import os
import pdb
import shutil

from typing import List, TypedDict, Union

from .app import App
from .constants import WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE, WORKFLOW_TEMPLATE_OPTION
from .docker.service.builder import ServiceBuilder
from .docker.workflow.mock_decorator import MockDecorator
from .docker.workflow.reverse_proxy_decorator import ReverseProxyDecorator
from .docker.workflow.builder import WorkflowBuilder
from .templates.factory import custom_files, maintained_files
from .workflow_command import WorkflowCommand

CORE_WORKFLOWS = [WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE]

class BuildOptions(TypedDict):
  builder_class: type 
  service_builder: ServiceBuilder
  template: WORKFLOW_TEMPLATE_OPTION
  workflow_decorators: List[Union[MockDecorator, ReverseProxyDecorator]]

class WorkflowCreateCommand(WorkflowCommand):

  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

    self.__env_vars: List[str] = kwargs.get('env') or []
    self.__force = not not kwargs.get('force')

  @property
  def workflow_templates_path(self):
    return self.build_workflow_templates_path(self.workflow_name)

  @property
  def env_vars(self):
    return self.__env_vars

  @property
  def force(self):
    return self.__force

  def build(self, **kwargs: BuildOptions):
    # Create workflow folder
    dest = os.path.join(self.workflow_path)
    if os.path.exists(dest) and self.force:
        shutil.rmtree(dest)

    if not os.path.exists(dest):
      os.makedirs(dest) 

    # Create workflow maintained compose file
    workflow_builder = self.__write_docker_compose_file(**kwargs)
    self.__copy_templates(workflow_builder, kwargs.get('template'))

    if not self.workflow_config.empty:
      self.workflow_config.write()

  def build_workflow_templates_path(self, workflow_name: str):
    return os.path.join(self.workflow_templates_root_dir, workflow_name)

  def __copy_templates(self, workflow_builder: WorkflowBuilder, template: WORKFLOW_TEMPLATE_OPTION = None):
    if not template:
      templates_path = self.workflow_templates_path
    else:
      templates_path = self.build_workflow_templates_path(template)

    if not os.path.exists(templates_path):
      return

    # Maintained files are files that will always be overwritten
    maintained_workflow_files = maintained_files(template or self.workflow_name, workflow_builder)
    self.copy_files(templates_path, maintained_workflow_files, self.workflow_path)

    # Custom files are files that may be modified by the user
    custom_workflow_files = custom_files(template or self.workflow_name, workflow_builder)
    self.copy_files_no_replace(templates_path, custom_workflow_files, self.workflow_path)

  def __write_docker_compose_file(self, **kwargs: BuildOptions):
    builder_class = kwargs.get('builder_class') or WorkflowBuilder
    service_builder = kwargs.get('service_builder') 
    if not service_builder:
      service_builder = ServiceBuilder(self.service_config)
      service_builder.load()
    workflow_decorators: List[Union[MockDecorator, ReverseProxyDecorator]] = kwargs.get('workflow_decorators')

    workflow_builder = builder_class(self.workflow_path, service_builder)
    workflow_builder.build_all()

    if isinstance(workflow_decorators, list):
      for workflow_decorator in workflow_decorators:
        workflow_decorator(workflow_builder).decorate()

    workflow_builder.write()
    workflow_builder.initialize_custom_file()

    return workflow_builder


from stoobly_orator.orm import has_many

from .base import Base
from .trace_alias import TraceAlias
from .trace_request import TraceRequest

class Trace(Base):
  __fillable__ = []

  @has_many
  def trace_aliases(self):
    return TraceAlias

  @has_many
  def trace_requests(self):
    return TraceRequest

  def clone(self):
    trace = self.create()
    if not trace:
      return

    for alias in self.trace_aliases:
      TraceAlias.create(
        name=alias.name,
        trace_id=trace.id,
        value=alias.value,
        value_inferred_type=alias.value_inferred_type,
      )

    return trace

def handle_deleting(trace):
  trace_aliaces = trace.trace_aliases

  for trace_alias in trace_aliaces:
    trace_alias.delete()

  trace_requests = trace.trace_requests
  for trace_request in trace_requests:
    trace_request.delete()

Trace.deleting(handle_deleting)


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, request, scenario
from stoobly_agent.lib.api.keys import ScenarioKey
from stoobly_agent.lib.orm.scenario import Scenario

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

@pytest.fixture(autouse=True, scope='module')
def settings():
  return reset()

class TestCreate():
  @pytest.fixture(scope='class')
  def scenario_name(self):
    return 'Login 1.0.0'

  @pytest.fixture(scope='class')
  def scenario_description(self):
    return 'Login flow for service A'

  @pytest.fixture(scope='class')
  def scenario_create_cli_results(self, runner: CliRunner, scenario_name: str, scenario_description: str):
    scenario_create_result = runner.invoke(scenario, ['create', '--without-headers', '--description', scenario_description, '--select', 'key', scenario_name])
    assert scenario_create_result.exit_code == 0

    return scenario_create_result

  @pytest.fixture(scope='class')
  def scenario_key(self, scenario_create_cli_results):
    output = scenario_create_cli_results.stdout
    return output.strip()

  def test_it_lists_scenario(self, runner: CliRunner, scenario_name: str, scenario_description: str, scenario_key: str):
    scenario_result = runner.invoke(scenario, ['list', '--without-headers', '--select', 'key'])
    assert scenario_result.exit_code == 0

    output = scenario_result.stdout
    _scenario_key = output.strip()

    _scenario = Scenario.find_by(uuid=ScenarioKey(_scenario_key).id)

    assert _scenario.key() == scenario_key
    assert _scenario.name == scenario_name
    assert _scenario.description == scenario_description

  def test_it_shows_scenario(self, runner: CliRunner, scenario_name: str, scenario_description: str, scenario_key: str):
    scenario_result = runner.invoke(scenario, ['show', '--without-headers', '--select', 'key', scenario_key])
    assert scenario_result.exit_code == 0

    output = scenario_result.stdout
    _scenario_key = output.strip()

    _scenario = Scenario.find_by(uuid=ScenarioKey(_scenario_key).id)

    assert _scenario.key() == scenario_key
    assert _scenario.name == scenario_name
    assert _scenario.description == scenario_description

  class TestWhenCreatingRequests():

    def test_it_increments_request_count(self, runner: CliRunner, scenario_key: str):
      requests_count = 2
      for i in range(requests_count):
        self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL, scenario_key)

      scenario_result = runner.invoke(scenario, ['show', '--without-headers', '--select', 'requests_count', scenario_key])
      assert scenario_result.exit_code == 0

      _requests_count = int(scenario_result.stdout.strip())
      assert _requests_count == requests_count
  
    def __record_request(self, runner: CliRunner, url: str, scenario_key: str):
      record_result = runner.invoke(record, ['--scenario-key', scenario_key, url])
      assert record_result.exit_code == 0

    class TestWhenDeletingRequests():

      def test_it_decrements_request_count(self, runner: CliRunner, scenario_key: str):
        requests_result = runner.invoke(request, ['list', '--scenario-key', scenario_key, '--without-headers', '--select', 'key'])
        assert requests_result.exit_code == 0

        keys = requests_result.stdout.strip().split("\n")
        for key in keys:
          self.__delete_request(runner, key)

        scenario_result = runner.invoke(scenario, ['show', '--without-headers', '--select', 'requests_count', scenario_key])
        assert scenario_result.exit_code == 0

        _requests_count = int(scenario_result.stdout.strip())
        assert _requests_count == 0

      def __delete_request(self, runner: CliRunner, request_key: str):
        delete_result = runner.invoke(request, ['delete', request_key])
        assert delete_result.exit_code == 0